package com.fundi.framework.common.displaylist;

/*
 *  Common Services Library - Framework plug-in.
 *  (C) Copyright Fundi Software 2011, 2016.  Fundi Confidential.
 *  All rights reserved.  Use, duplication or disclosure restricted. 
 */

import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.Marshaller;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.bind.annotation.XmlTransient;

@XmlRootElement
public class ListLayoutV1 extends ListLayout {

	//Eye catchers.
	public static final String EC_COPYRIGHT = "(C) Copyright 2011, 2016 Fundi Software.  Fundi Confidential."; //$NON-NLS-1$
	public static final String EC_APAR = "3.1.1.00_00014"; //$NON-NLS-1$
	public static final String EC_FMID = "H30SA10"; //$NON-NLS-1$
	public static final String EC_PID = "5655-CSL"; //$NON-NLS-1$
	
	private static final long serialVersionUID = 1L;
	private static final ArrayList<String> excludeValues = new ArrayList<String>(Arrays.asList(new String[]{DisplayList.nothingLabel, "", null}));
	
	private Map<String, ColumnLayout>	columns = new HashMap<String, ColumnLayout>();
	private Integer						version = 1;	//leaving version 0 to represent the serialized data
	
	public ListLayoutV1() {
		super();
	}

	public ListLayoutV1(String context) {
		super(context);
	}
	
	public ListLayoutV1(ListLayout legacyLayout) throws IllegalArgumentException {
		ListFieldDefn	defn;
		int				keySeq;

		context = legacyLayout.getContext();
		setFieldSequence(clean(legacyLayout.getFieldSequence()));
		setKeyFields(clean(legacyLayout.getKeyFields()));
		for (String field: getFieldSequence()) {
			if (field != null && !field.isEmpty()) {
				defn = new ListFieldDefn();
				defn.setName(field);
				keySeq = getKeyFields().indexOf(field);
				if (keySeq != -1) {
					defn.setKey(true);
					defn.setKey(keySeq);
				}
				columns.put(field,  new ColumnLayout(defn));
			}
		}
		setAscending(legacyLayout.isAscending());
		setHidingBlankColumns(legacyLayout.isHidingBlankColumns());
		setHidingZeroValueColumns(legacyLayout.isHidingZeroValueColumns());
		setHidingSameValueColumns(legacyLayout.isHidingSameValueColumns());
	}
	
	public boolean isEmpty() {
		return getFieldSequence().isEmpty();	//based on the current situation where excludedfields do not get included into the fieldsequence
	}
	
	public static ArrayList<String> clean(ArrayList<String> input) {
		ArrayList<String> 	result;
		
		result = new ArrayList<String>(input.size());
		for (String field: input) {
			if (!excludeValues.contains(field) && !result.contains(field)) {	//MA:Added uniqueness to the clean function as this is causing problems in resource type All for GPL where there are aliased fields
				result.add(field);
			}
		}
		
		return result;
	}
	
	public synchronized void set(DisplayList list) throws IllegalArgumentException {
		setFieldSequence(list.getAllFields(), list);
		setKeyFields(list.getKeyFields(), list);
		setAscending(list.isAscending());
		setHidingBlankColumns(list.isHidingBlankColumns());
		setHidingZeroValueColumns(list.isHidingZeroValueColumns());
		setHidingSameValueColumns(list.isHidingSameValueColumns());
	}

	public synchronized void setDefaults(DisplayList list) throws IllegalArgumentException {
		getFieldSequence().clear();
		setAscending(true);
		setHidingBlankColumns(false);
		setHidingZeroValueColumns(false);
		setHidingSameValueColumns(false);
		setFieldSequence(list.getDefaultFieldSequence(), list);
		setKeyFields(list.getDefaultKeyFieldLabels(), list);
	}
	
	public void setFieldSequence(ArrayList<String> orderedFields, DisplayList list) throws IllegalArgumentException {
		ArrayList<String>	working;
		
		working = clean(orderedFields);
		ensureColumnLayoutsExist(working, list);
		setFieldSequence(working);
	}
	
	public void ensureColumnLayoutsExist(ArrayList<String> orderedFields, DisplayList list) throws IllegalArgumentException {
		ColumnLayout	columnLayout;
		final boolean	MUST_BE_IN_ALL_ROWS = true;
		
		for (String field: orderedFields) {
			if (field != null && !field.isEmpty() && columns.get(field) == null) {
				ListFieldDefn defn = list.getFieldDefn(field);
				if (defn == null) {
					columnLayout = new ColumnLayout(field);
				} else {
					columnLayout = new ColumnLayout(defn);
				}
				columns.put(field, columnLayout);
				if (!excludeValues.contains(field) && !getFieldSequence().contains(field)) {	//'nothingLabel' needs to go in the column layouts but not the fields sequence
					getFieldSequence().add(field);
				}
			}
		}
		for (ListFieldDefn defn: list.getFieldDefns(!MUST_BE_IN_ALL_ROWS)) {	//MA: need to work through and pick up information from all available fields - else those that are server hidden by the common code might show up in the ListLayoutConfigurationDialog
			columnLayout = columns.get(defn.getLabel());
			if (columnLayout != null) {
				columnLayout.setHiddenByServer(defn.isHiddenByServer());
			}
		}
	}

	public void setKeyFields(ArrayList<String> orderedFields, DisplayList list) throws IllegalArgumentException {
		ArrayList<String>	working;
		
		working = clean(orderedFields);
		ensureColumnLayoutsExist(working, list);
		setKeyFields(working);
	}
	
	public void moveTop(ArrayList<String> selection, DisplayList list) {
		ArrayList<String>	working,
							fieldSequence;

		working = clean(selection);
		fieldSequence = getFieldSequence();
		for (String label: fieldSequence) {
			if (!working.contains(label)) {
				working.add(label);
			}
		}
		setFieldSequence(working, list);
	}

	public void moveUp(ArrayList<String> selection, DisplayList list) {
		ArrayList<String>	fieldSequence;
		boolean				prevInSelection;
		ArrayList<String>	working;
		
		working = clean(selection);
		fieldSequence = getFieldSequence();
		prevInSelection = true;	//don't allow the first element to move even if it is checked - equivalent to checking for index != 0
		for (int index = 0; index < fieldSequence.size(); index++) {
			if (working.contains(fieldSequence.get(index))) {
				if (!prevInSelection) {
					Collections.swap(fieldSequence, index - 1, index);
				}
				//here we leave prevInSelection as it is because we've just swapped the elements if it was false 
				//- so it should stay false for the next iteration, and otherwise it's true which is what it needs 
				//to be set to for the next iteration in this case.
			} else {
				prevInSelection = false;
			}
		}
	}

	public void moveDown(ArrayList<String> selection, DisplayList list) {
		ArrayList<String>	fieldSequence;
		boolean				prevInSelection;
		ArrayList<String>	working;
		
		working = clean(selection);
		fieldSequence = getFieldSequence();
		prevInSelection = true;	//don't allow the first element to move even if it is checked - equivalent to checking for index != 0
		for (int index = fieldSequence.size() - 1; index >=0; index--) {
			if (working.contains(fieldSequence.get(index))) {
				if (!prevInSelection) {
					Collections.swap(fieldSequence, index + 1, index);
				}
				//here we leave prevInSelection as it is because we've just swapped the elements if it was false 
				//- so it should stay false for the next iteration, and otherwise it's true which is what it needs 
				//to be set to for the next iteration in this case.
			} else {
				prevInSelection = false;
			}
		}
	}
	
	public void moveRelative(String field, ArrayList<String> sequence) {
		ArrayList<String>	fieldSequence,
							working;
		int					oldIndex,
							newIndex;
		String				placementItem;
		ArrayList<String>	cleanSequence;
		
		cleanSequence = clean(sequence);
		//reconstruct the old ordering of just the sequenced columns
		fieldSequence = getFieldSequence();
		working = new ArrayList<String>();
		for (String workingField: fieldSequence) {
			if (cleanSequence.contains(workingField)) {
				working.add(workingField);
			}
		}
		oldIndex = working.indexOf(field);
		newIndex = cleanSequence.indexOf(field);
		if (oldIndex > newIndex) {
//			//moved up in the list
//			fieldSequence.remove(field);
//			//the item it has moved in front of is
//			placementItem = cleanSequence.get(newIndex + 1);
//			//so add it at this items position to insert it here
//			fieldSequence.add(fieldSequence.indexOf(placementItem), field);
		} else if (oldIndex < newIndex) {
			//moved down in the list
			fieldSequence.remove(field);
			//the item it has moved after is
			placementItem = cleanSequence.get(newIndex - 1);
			//so add it after this items position to insert it here
			fieldSequence.add(fieldSequence.indexOf(placementItem) + 1, field);
		}
	}

	public void moveBottom(ArrayList<String> selection, DisplayList list) {
		ArrayList<String>	fieldSequence;
		Iterator<String>	selectionIterator;
		String				label;
		ArrayList<String>	working;
		
		working = clean(selection);
		fieldSequence = getFieldSequence();
		selectionIterator = working.iterator();
		while (selectionIterator.hasNext()) {
			label = selectionIterator.next();
			if (label != null && !label.isEmpty()) {
				fieldSequence.remove(label);
				fieldSequence.add(label);
			}
		}
	}

	@Override
	public String toString() {
		JAXBContext		xmlContext;
		Marshaller		m;
		StringWriter 	writer;
		String			result;
		
		result = null;
		try {
			xmlContext = JAXBContext.newInstance(ListLayoutV1.class);
			m = xmlContext.createMarshaller();
			m.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
			writer = new StringWriter();
			m.marshal(this, writer);
			result = writer.toString();
			writer.close();
		} catch (Throwable e) {
			e.printStackTrace();
		}
		
		return result;
	}

	/**
	 * the list of all known columns built off the allFields index and therefore including even those not present in the current list
	 * @return
	 */
	public List<ColumnLayout> getAllColumns() {
		List<ColumnLayout>	result;
		
		result = new ArrayList<ColumnLayout>(getFieldSequence().size());
		for (String field: getFieldSequence()) {
			result.add(columns.get(field));
		}
				
		return result;
	}

	@XmlTransient
	public void setColumnList(List<ColumnLayout> columns) {
		this.columns.clear();
		this.getFieldSequence().clear();
		for (ColumnLayout column: columns) {
			if (column != null) {
				this.columns.put(column.getLabel(), column);
				this.getFieldSequence().add(column.getLabel());
			}
		}
	}
	
	public ColumnLayout getColumnLayout(String label) {
		return columns.get(label);
	}

	public int getSortIndex(String label) {
		return getKeyFields().indexOf(label);
	}

	public int getVersion() {
		return version;
	}

	public void setVersion(int version) {
		this.version = version;
	}

	public Map<String, ColumnLayout> getColumns() {
		return columns;
	}

	public void setColumns(Map<String, ColumnLayout> columns) {
		this.columns = columns;
	}
	
	public ArrayList<String> getAllVisibleFields(DisplayList list) {
		return getAllVisibleFields(list, false);
	}
	
	public ArrayList<String> getAllVisibleFields(DisplayList list, boolean strict) {
		ArrayList<String>		columnLabels;
		ColumnLayout			column;
		
		columnLabels = new ArrayList<String>(columns.size());
		for (String label: list.getAllVisibleFields(strict)) {
			column = getColumnLayout(label);
			if (column == null || ((strict && !column.isHiddenByServer()) || (!strict && column.isVisible()))) {
				columnLabels.add(label);
			}
		}

		return columnLabels;
	}
}
